import asyncio
import logging
from typing import Dict, List, Union, Any

import azure.cosmos.exceptions as exceptions
from fastapi import APIRouter, Depends, HTTPException, Query
from fastapi.responses import JSONResponse
from pydantic import BaseModel

from ..cache import Cache
from ..database.mycosmosdb import callConnect
from ..security import get_current_active_user
from .utils import read_item

router = APIRouter()


DataItem = Dict[str, Union[str, Dict[str, str], List[str]]]
CarbonDataItem = Dict[str, Any] # TODO : check if any restriction required 

class PostData(BaseModel):
    oid: str
    email: str
    metric: str
    year: str
    quarter: str
    data: Union[List[DataItem], DataItem]

class CarbonPostData(PostData):
    data: Union[List[CarbonDataItem], CarbonDataItem]



def build_metrics_data(metric_label: str, year, quarter, data, diversity=False):
    """
    Build metrics data for a given metric, year, quarter and data.
    """
    if isinstance(data, list):
        quarter_metrics = {"label": "1", "value": data}
    else:
        quarter_metrics = {"label": "1", "value": [data]}
    if diversity:
        quarter_data = {"label": quarter, "value": [data]}
    else:
        quarter_data = {"label": quarter, "metrics": [quarter_metrics]}
    year_data = {"label": year, "value": [quarter_data]}
    metric_data = {"label": metric_label, "metrics": [year_data]}
    return metric_data, year_data, quarter_data


def update_measurement(
    measurement, metric_label: str, year: str, quarter: str, data, diversity=False
):
    """
    Update measurement for a given metric, year, quarter and data.
    """
    metric_data, year_data, quarter_data = build_metrics_data(
        metric_label=metric_label,
        year=year,
        quarter=quarter,
        data=data,
        diversity=diversity,
    )

    # check if given metric exists, for eg: Environmental Impact, etc
    result = next((x for x in measurement if x["label"] == metric_label), None)

    if result is None:
        # if metric with given label doesn't exist -> append metric data for the given timeperiod
        return measurement + [metric_data]

    # else append data for the given timeperiod in the existing metric
    # retrieve other metrics data
    other_metrics = [x for x in measurement if x["label"] != metric_label]

    # find data for given year
    result_year = next((x for x in result["metrics"] if x["label"] == year), None)

    if result_year:
        other_year_results = [x for x in result["metrics"] if x["label"] != year]

        # if data with given year exists -> append quarter data with
        # given quarter and metrics
        other_quarter_results = [
            x for x in result_year["value"] if x["label"] != quarter
        ]

        result_year["value"] = other_quarter_results + [quarter_data]
        result["metrics"] = other_year_results + [result_year]
    else:
        # if data with given year doesn't exist
        # -> append year data with given quarter and metrics
        result["metrics"].append(year_data)

    return other_metrics + [result]

@router.post("/carbon-measure")
async def post_carbon_measure(data: CarbonPostData):
    """
    Post carbon measure data for a given user.
    """
    try:
        try:
            db_result = read_item(callConnect.container7, data.oid, data.email)
        except exceptions.CosmosHttpResponseError as e:
            logging.error("This is an error message for /carbonmeasure")
            raise HTTPException(status_code=500, detail="Internal Server Error")

        if "carbonmeasure" not in db_result:
            raise HTTPException(
                status_code=404, detail="Data not found in the database"
            )
        carbon_measure = db_result["carbonmeasure"]
        user = {"id": db_result["id"], "email": db_result["email"], "carbonmeasure": []}

        updated_measurement = update_measurement(
            measurement=carbon_measure,
            metric_label=data.metric,
            year=data.year,
            quarter=data.quarter,
            data=data.data,
        )
        user["carbonmeasure"] = updated_measurement
        callConnect.container7.upsert_item(body=user)
        return JSONResponse(content="Data added successfully", status_code=200)

    except HTTPException as e:
        return JSONResponse(status_code=e.status_code, content={"message": e.detail})


@router.post("/custom-measure")
async def post_custom_measure(data: PostData):
    """
    Post custom measure data for a given user.
    """
    try:
        try:
            db_result = read_item(callConnect.container15, data.oid, data.email)
        except exceptions.CosmosHttpResponseError as e:
            logging.error("This is an error message for /custommeasure")
            raise HTTPException(status_code=500, detail="Internal Server Error")

        if "custommeasure" not in db_result:
            raise HTTPException(
                status_code=404, detail="Data not found in the database"
            )
        custom_measure = db_result["custommeasure"]
        user = {"id": db_result["id"], "email": db_result["email"], "custommeasure": []}

        updated_measurement = update_measurement(
            measurement=custom_measure,
            metric_label=data.metric,
            year=data.year,
            quarter=data.quarter,
            data=data.data,
        )
        user["custommeasure"] = updated_measurement
        callConnect.container15.upsert_item(body=user)
        return JSONResponse(content="Data added successfully", status_code=200)

    except HTTPException as e:
        return JSONResponse(status_code=e.status_code, content={"message": e.detail})


@router.post("/diversity-measure")
async def post_diversity_measure(data: PostData):
    """
    Post diversity measure data for a given user.
    """
    try:
        try:
            db_result = read_item(callConnect.container8, data.oid, data.email)
        except exceptions.CosmosHttpResponseError as e:
            logging.error("This is an error message for /diversitymeasure")
            raise HTTPException(status_code=500, detail="Internal Server Error")

        if "diversitymeasure" not in db_result:
            raise HTTPException(
                status_code=404, detail="Data not found in the database"
            )
        custom_measure = db_result["diversitymeasure"]
        user = {
            "id": db_result["id"],
            "email": db_result["email"],
            "diversitymeasure": [],
        }
        updated_measurement = update_measurement(
            measurement=custom_measure,
            metric_label=data.metric,
            year=data.year,
            quarter=data.quarter,
            data=data.data,
            diversity=True,
        )
        user["diversitymeasure"] = updated_measurement
        callConnect.container8.upsert_item(body=user)
        return JSONResponse(content="Data added successfully", status_code=200)

    except HTTPException as e:
        return JSONResponse(status_code=e.status_code, content={"message": e.detail})


class User(BaseModel):
    oid: str
    email: str
    city: str


async def query_custom_measurements_async(container, user_ids):
    try:
        formatted_user_ids = [f'"{user_id}"' for user_id in user_ids]
        # Construct SQL-like query to retrieve custom measurement data for specified user IDs
        query = f"SELECT * FROM c WHERE c.id in ({','.join(formatted_user_ids)})"

        # Execute query
        results = container.query_items(query=query, enable_cross_partition_query=True)

        return results
    except exceptions.CosmosHttpResponseError as e:
        logging.error(f"Error querying custom measurements: {e}")
        raise


# Create an instance of the Cache class with a TTL of 5 minutes
cache = Cache(ttl=300)


@router.get("")
async def get_measurements(
    investor_oid: str = Query(..., description="The OID of the investor user"),
    investor_email: str = Query(..., description="The email of the investor user"),
    investor_city: str = Query(..., description="The city of the investor user"),
    authenticated_user: User = Depends(get_current_active_user),
):
    """
    This endpoint retrieves all measurements for a given user of type fund/investor.

    """
    data = {
        "oid": investor_oid,
        "email": investor_email,
        "city": investor_city,
        "measurements": {},
    }

    try:
        user_record = read_item(callConnect.container, investor_oid, investor_city)
        if not user_record:
            return JSONResponse(
                content={"message": "User data not found in the database"},
                status_code=404,
            )

        funds = user_record.get("funds", [])

        if not funds:
            return JSONResponse(
                content={"message": "No funds found for the user"}, status_code=404
            )

        # Check if the data is already in the cache
        cached_data = cache.get(investor_oid)
        if cached_data:
            return JSONResponse(content=cached_data, status_code=200)

        user_ids = []
        for fund in funds:
            for company in fund.get("companies", []):
                if company["oid"] not in user_ids:
                    user_ids.append(company["oid"])
                    data["measurements"][company["oid"]] = {
                        "oid": company["oid"],
                        "email": company["email"],
                        "c_name": company["id"],
                        "carbonmeasure": [],
                        "custommeasure": [],
                        "diversitymeasure": [],
                    }
        (
            carbon_measurements,
            custom_measurements,
            diversity_measurements,
        ) = await asyncio.gather(
            query_custom_measurements_async(callConnect.container7, user_ids),
            query_custom_measurements_async(callConnect.container15, user_ids),
            query_custom_measurements_async(callConnect.container8, user_ids),
        )

        async def process_measurements(measurements, key):
            for item in measurements:
                data["measurements"][item["id"]][key] = item[key]

        await asyncio.gather(
            process_measurements(carbon_measurements, "carbonmeasure"),
            process_measurements(custom_measurements, "custommeasure"),
            process_measurements(diversity_measurements, "diversitymeasure"),
        )

        # Store the result in cache before returning
        cache.set(investor_oid, data)

        return JSONResponse(content=data, status_code=200)

    except exceptions.CosmosResourceNotFoundError:
        return JSONResponse(
            content={"message": "User data not found in the database"}, status_code=404
        )

    except exceptions.CosmosHttpResponseError as e:
        logging.error(e)
        raise HTTPException(status_code=500, detail="Internal Server Error")
